/* zeelib.c : [De]Compression using zlib
 * One object: ZeeLib, containing two objects: Compressor and Inflator
 * See http://zlib.net/
 * ZLib objects are Stream/Pipe/filter like.
 *   write() doesn't flush and is very "chunky"
 *   read() blocks, unless told not to
 * In general:
 *   Flushed on close(), Inflators self close when they see the end of the
 *     compressed data.
 *   Closing a closed ZLib is fine.
 * Thread safety:
 *   Compressors:  Writes are thread safe, thus Compressor reads are also.
 *     It makes sense to support compressing packets from multiple threads.
 *     It DOES NOT make sense to have mutiple consumers.
 *   Inflators:  While multiple consumers of uncompressed data *might* make
 *     sense (if packetized), multiple producers do not.  However, since
 *     I/you have no control over read data size, you can't just read a
 *     packet.  So, doesn't have to be thread safe.
 * Handles gzip encoding (both directions).
 * Uses zlib shared library: zlib1.dll
 * 
 * Data structure:  List of [de]compressed data chunks.  As the ZLib object
 * is written to, the data is munged into a buffer, when the buffer is full,
 * it is added to the list and another buffer is allocated.  Read() pulls
 * from the list, until the object is closed, then the buffer.
 * While this is pretty hokey (it is much easier to just let the Data grow
 * as needed), it makes error recovery and thread safety easier to deal
 * with.
 * 
 * Compile multi-threaded (MSVC properties/C/Code Generation/Runtime library)
 * Linked against zdll.lib (zlib exports), zkl.lib (zkl exports) and
 *    msvcrt.lib.
 *    In MSVC properties/Linker/Input, set:
 *    - "Addition Dependencies" to:
 *		zdll.lib ..\..\VM\Debug\zkl.lib msvcrt.lib
 *       (modulo the release and where zkl.lib actually is).
 *    - "Ignore All Default Libraries" to "Yes (/NODEFAULTLIB)".  This
 *      creates a much smaller DLL and is why you need to link against
 *      msvcrt.lib
 * Creates: zeelib.lib (exports), zeelib.dll
 * 
 * Linux zlib package name: zlib1g-dev, link against /usr/lib/libz.so (-lz)
 * 
 * Copyright (c) 2006, 2007, 2008 Craig Durland craigd@zenkinetic.com
 * License: The zlib/libpng License,
 * 	http://www.opensource.org/licenses/zlib-license.php
 */

#define _CRT_SECURE_NO_WARNINGS		// VC++ 2008

#include <stdio.h>
#include <string.h>

#include "zlib.h"		// this code written against version 1.2.3

#ifdef __GNUC__
   #define BYTE_DEFINED	// GCC doesn't like 2 typedefs for Byte
#endif
#include "zklObject.h"
#include "zklAtomic.h"
#include "zklData.h"
#include "zklList.h"
#include "zklNumber.h"
#include "zklString.h"
#include "zklUtil.h"
#include "zklImports.h"


static Instance *calcAdler32(Instance *,pArglist,pVM);
static Instance *calcCRC32(Instance *,pArglist,pVM);
static size_t    zlibLen(Instance *);

#define CHUNK		    16384

#define INFLATOR		0
#define COMPRESSOR		1

#define Z_DEFAULT_WINDOW_BITS  15	// version 1.2.3
#define Z_DEFAULT_MEM_LEVEL	8

static ZKL_Object CompressorObject, InflatorObject;

typedef struct		// inherits from Instance
{
   Instance	instance;

   CAtomicInt	zOpen;
   SpinLock	zLock;
   int		zType;		// COMPRESSOR or INFLATOR
   z_stream	zStream;
   Instance    *zData;		// buffer, a Data
   Instance    *blockList;	// List of buffers
} ZLib;

#define Z_OPEN(i)	( ((ZLib *)i)->zOpen )
#define	Z_LOCK(i)	( &(((ZLib *)i)->zLock) )
#define Z_DATA(i)	( ((ZLib *)i)->zData )
#define BLOCK_LIST(i)	( ((ZLib *)i)->blockList )
#define Z_TYPE(i)	( ((ZLib *)i)->zType )
#define Z_STREAM(i)	( &(((ZLib *)i)->zStream) )

static void _broken(Instance *self);

static Instance *zlibCreate(int compressor, pVM vm)
{
   ZLib	       *zlib;

   zlib = compressor ? 
	    (ZLib *)instanceAllocate(sizeof(ZLib),&CompressorObject,1,vm) :
	    (ZLib *)instanceAllocate(sizeof(ZLib),&InflatorObject,1,vm);
   CAI_INIT(&zlib->zOpen);
   Z_TYPE(zlib)	    = compressor;
   spinLockInit(Z_LOCK(zlib));
   Z_DATA(zlib)	    = 0;
   BLOCK_LIST(zlib) = 0;
   return addToCollectables((Instance *)zlib,I_OWNED,vm);
}

static z_stream *getStream(Instance *self, pVM vm)
{
   z_stream *stream = Z_STREAM(self);
   if (!Z_OPEN(self))
   {
      spinLockRelease(Z_LOCK(self));	// probably redundant
      vmThrow(vm,E_IO_ERROR, "ZLib not open, can't write to it");
   }
   return stream;
}

    /* Another way to do this would be to create Z_DATA & BLOCK_LIST
     * invisible and make them owned only when giving them to somebody else
     * or at free. Then, don't need to mark because GC doesn't see them.
     */
static void magicMarker(Instance *self)
{
   instanceMark(Z_DATA(self));
   instanceMark(BLOCK_LIST(self));
}

    // clean up z_stream, GC will take care of data and blocklist
static int freeMe(Instance *self)
{
   int       ret;
   z_stream *stream = Z_STREAM(self);

   if (!Z_OPEN(self)) return 1;
   ret = Z_TYPE(self) ? deflateEnd(stream) : inflateEnd(stream);
   return 1;
}

static Instance *_open(int compressor, pArglist arglist, pVM vm)
{
   Instance    *zlib;
   int		ret;
   z_stream    *stream;

   zlib		    = zlibCreate(compressor,vm);
   Z_DATA(zlib)	    = 0;
   BLOCK_LIST(zlib) = 0;
   Z_DATA(zlib)     = dataCreate(CHUNK,I_OWNED,vm);	// can throw
   stream	    = Z_STREAM(zlib);
   stream->zalloc   = Z_NULL;
   stream->zfree    = Z_NULL;
   stream->opaque   = Z_NULL;
   if (compressor)
   {
      int b = arglistTryToGetBool(arglist,0,0,"zeelib.open",vm);
      int gzipWrapper = b * 16;		// True-->16, False (default) -->0
      BLOCK_LIST(zlib) = listCreate(1,1,I_OWNED,vm);

      ret = deflateInit2(stream, Z_DEFAULT_COMPRESSION, Z_DEFLATED,
	      Z_DEFAULT_WINDOW_BITS + gzipWrapper,
	      Z_DEFAULT_MEM_LEVEL, Z_DEFAULT_STRATEGY);
   }
   else		// uncompressing
   {
      stream->avail_in = 0;
      stream->next_in  = Z_NULL;
      BLOCK_LIST(zlib) = listCreate(1,0,I_OWNED,vm);
      ret = inflateInit2(stream,Z_DEFAULT_WINDOW_BITS + 32);
   }
   if (ret != Z_OK)
   {
      char buffer[100];
      sprintf(buffer,"ZLib.*.open: %s",stream->msg);
      vmThrow(vm,E_IO_ERROR,buffer);
   }

   Z_OPEN(zlib) = 1;
   return zlib;
}

static void fencedUnlocker(Fence *fence, Instance *e)
   { spinLockRelease(Z_LOCK(fence->i)); }

static void _packBlewAFuse(Fence *fence, Instance *e)
{
   Instance *self = fence->i;
   _broken(self);
   spinLockRelease(Z_LOCK(self));	// this works even if not locked
}
    // Self has been locked
static void _packItIn(Instance *self, unsigned len, unsigned gapSize, pVM vm)
{
   z_stream *stream = Z_STREAM(self);
   Instance *data   = Z_DATA(self);
   unsigned  have;
   Fence     fence;

   have = gapSize - stream->avail_out;
   dataSetSize(data,len + have);
   if (dataLen(data) == CHUNK)	// block is full, get another block
   {
      vmSetFence(vm,&fence,_packBlewAFuse,self);
         listAppend(BLOCK_LIST(self),data,vm);
	 Z_DATA(self) = dataCreate(CHUNK,I_OWNED,vm);
      vmRemoveFence(&fence,0);
   }
   asyncWriteHappened(self);	// maybe, maybe not
}

    // Write raw data to a Compressor.
    // numBytes == 0 --> end of input
    // Multiple producers is OK
    // YOU must lock so that this is thread safe
static void _compress(Instance *self, void *bytes, size_t numBytes, pVM vm)
{
   Instance *data;
   int	     ret, flush;
   unsigned  len, gapSize;
   z_stream *stream;

   flush	    = (numBytes == 0) ? Z_FINISH : Z_NO_FLUSH;
   stream	    = getStream(self,vm);
   stream->avail_in = (uInt)numBytes;
   stream->next_in  = bytes;
   do
   {
      data    = Z_DATA(self);
      len     = (unsigned)dataLen(data);
      gapSize = CHUNK - len;

      stream->avail_out = gapSize;
      stream->next_out  = (Byte *)dataWillFill(data) + len;
      ret = deflate(stream,flush);	// no bad return value
      _packItIn(self,len,gapSize,vm);
   } while(stream->avail_out == 0);
}

    // *.close(): Close while writing == cluster fuck but no crash
    // Self has been locked
static void _close(Instance *self,pVM vm)
{
   int	     ret;
   z_stream *stream = Z_STREAM(self);

   if (!Z_OPEN(self)) ret = Z_OK;   	// inflateEnd() twice isn't good
   else
   {
      if (Z_TYPE(self) == COMPRESSOR)
      {
	 _compress(self,0,0,vm);
	 ret = deflateEnd(stream);
      }
      else ret = inflateEnd(stream);	// inflating

      CAI_ZERO(&Z_OPEN(self));		// closed
      asyncClosed(self);
   }

   if (ret != Z_OK)
   {
      char buffer[100];
      sprintf(buffer,"ZLib.*.close: %s",stream->msg);
      spinLockRelease(Z_LOCK(self));
      vmThrow(vm,E_IO_ERROR,buffer);
   }
}
static Instance *ZLib_close(Instance *self, pArglist arglist, pVM vm)
{
   spinLockAcquire(Z_LOCK(self));
      _close(self,vm);
   spinLockRelease(Z_LOCK(self));
   return self;
}

    // Pretty much the same as close
    // Lock released, just because I can
static void _broken(Instance *self)
{
   int	     ret;
   z_stream *stream = Z_STREAM(self);
   
   CAI_ZERO(&Z_OPEN(self));
   if (Z_TYPE(self) == COMPRESSOR)
   {
      ret = deflateEnd(stream);
   }
   else		// inflating
   {
      ret = inflateEnd(stream);
   }
   spinLockRelease(Z_LOCK(self));
   asyncClosed(self);
}

    /* Write compressed data to an Inflator.
     * Multiple producers == cluster fuck.
     * DOES NOT need to be thread safe
     */
static void _uncompress(Instance *self, void *bytes, size_t numBytes, pVM vm)
{
   Instance    *data;
   int		ret;
   z_stream    *stream;
   unsigned	len, gapSize;

//   spinLockAcquire(Z_LOCK(self));
      stream	       = getStream(self,vm);
      stream->avail_in = (uInt)numBytes;
      stream->next_in  = bytes;
      do
      {
	 data    = Z_DATA(self);
	 len     = (unsigned)dataLen(data);
	 gapSize = CHUNK - len;

	 stream->avail_out = gapSize;
	 stream->next_out  = (Byte *)dataWillFill(data) + len;
	 ret = inflate(stream, Z_NO_FLUSH);
	 if (ret < 0 || ret == Z_NEED_DICT)	// this is rather rank
	 {
	    char	buffer[100];
	    sprintf(buffer,"ZLib.Inflator.write: %s",stream->msg);
	    _broken(self);	// releases lock
	    vmThrow(vm,E_IO_ERROR,buffer);
	 }
	 _packItIn(self,len,gapSize,vm);
      } while (stream->avail_out == 0);
      if (ret == Z_STREAM_END) _close(self,vm);
//   spinLockRelease(Z_LOCK(self));
}

    /* Is there data available for reading? Only if:
     *   - there are full blocks available
     *   - closed and something is in buffer
     * Don't read buffer because it is in transition.
     */
static int _gotData(void *self)
{
   size_t len = listLen(BLOCK_LIST(self),NoVM);	// thread safe

   if (len)	      return A_OK;  // got data
   if (!Z_OPEN(self)) return A_OK;  // closed, maybe has data
   return	      A_FAIL;	    // still open but no data [yet]
}

static int _isClosed(void *self) { return !Z_OPEN(self) ? A_OK : A_FAIL; }

    /* *.read([timeout]) --> one block of data
     * Multiple consumers don't make sense but needs to be thread safe for
     * Compressor.write(). blocklist isn't thread safe for Inflators.
     */
static Instance *_zread(Instance *self,pArglist arglist,int throw,pVM vm)
{
   Instance *data;
   Instance *timeout = arglistTryToGet(arglist,0);
   Instance *list    = BLOCK_LIST(self);
   int	     s;

   	// never opened or bled dry
   if (!Z_DATA(self))
   {
      if (throw) vmThrowTheEnd(vm);
      return 0;
   }

   s = asyncRead(self,_gotData,(void *)self,timeout,vm);
   if (s < A_OK) return Void;		// timed out or interupted

   	// closed or data is in list
   data = listPop(list,0,vm);	// returns 0 if empty
   if (data) { instanceIsOrphan(data); return data; }

	/* closed, data might hold something.
	 * Wrong case:  Multiple threads read an open zlib with one chunk at
	 * same time, one thread will get here with no chunks and still
	 * open.  DON'T nuke zData in that case!
	 * This isn't a true solution but they wrote bad zkl code.
	 */
   spinLockAcquire(Z_LOCK(self));
      data = Z_DATA(self);
      if (!Z_OPEN(self)) Z_DATA(self) = 0;
   spinLockRelease(Z_LOCK(self));

   if (data)  return data;
   if (throw) vmThrowTheEnd(vm);
   return 0;
}

static Instance *ZLib_read(Instance *self,pArglist arglist,pVM vm)
  { return _zread(self,arglist,1,vm); }

    // *.drain([timeout])
    // Wait until close and return all data.
    // Thread safe
static Instance *ZLib_drain(Instance *self,pArglist arglist,pVM vm)
{
   Fence	fence;
   Instance    *data;
   Instance    *timeout = arglistTryToGet(arglist,0);
   Instance    *list    = BLOCK_LIST(self);
   Instance    *buffer;
   size_t	n;

   	// never opened or bled dry
   if (!Z_DATA(self)) return getEmptyKData();

	// wait until closed
   if (asyncRead(self,_isClosed,(void *)self,timeout,vm) < A_OK)
	return Void;		// timed out or interupted

// a pump would work nicely here
   spinLockAcquire(Z_LOCK(self));
   vmSetFence(vm,&fence,fencedUnlocker,self);
      n    = listLen(BLOCK_LIST(self),NoVM) * CHUNK + dataLen(Z_DATA(self));
      data = dataCreate(n,I_OWNED,vm);
      fence.i1 = data;		// protect added contents from GC
      while (1)
      {
	 buffer = listPop(list,0,vm);	// returns 0 if empty
	 if (!buffer) break;
	 dataAppend(data,buffer,vm);	// can GC. Won't throw
      }
      buffer = Z_DATA(self);
      dataAppend(data,buffer,vm);	// won't GC (Data is big enough)
      Z_DATA(self) = 0;			// keep marking Z_DATA until copied
   vmRemoveFence(&fence,1);
   return data;
}

static Instance *_zGetter(Instance *self,size_t n,void *X,size_t _,pVM vm)
   { return _zread(self,NoArglist,0,vm); }

    // *.pump(sink,actions)
    // Compressor.pump -->
    // Inflator.pump   -->
static Instance *ZLib_pump(Instance *self, pArglist a, pVM vm)
   { return pump(self,0x00,_zGetter,0,a,0,vm); }

    // *.walker
static Instance *ZLib_walker(Instance *self, pArglist a, pVM vm)
   { return walkerCreate(self,_zGetter,0,0,vm); }

static size_t zlibLen(Instance *self)
{
   size_t numBlocks, len;

   	// never opened or bled dry
   if (!Z_DATA(self)) return 0;

   spinLockAcquire(Z_LOCK(self));
      numBlocks = listLen(BLOCK_LIST(self),NoVM);
      len 	= dataLen(Z_DATA(self));
   spinLockRelease(Z_LOCK(self));
   return numBlocks * CHUNK + len;
}

    // *.len()
static Instance *ZLib_len(Instance *self, pArglist a, pVM vm)
   { return intCreate(zlibLen(self),vm); }

/* ******************************************************************** */
/* ************************ Compressor Methods ************************ */
/* ******************************************************************** */

    // Compressor.create([add gzip wrapper])
    // Compressor.open
    // NOTE! if you open an existing Compressor, you still get a new one
static Instance *Compressor_open(Instance *self,pArglist arglist,pVM vm)
   { return _open(COMPRESSOR,arglist,vm); }

static void compressorWrite(Instance *self,Instance *x,pVM vm)
{
   void	  *ptr;
   size_t  size;
   Byte	   c;

   switch (TYPEO(x))
   {
      case DataType:   ptr = dataText(x,&size);			break;
      case StringType: ptr = stringText(x); size = strlen(ptr); break;
      case ListType: case TupleType:
      {
	 size_t		i;
	 Instance      *t;
	 for (i = 0; (t = listGet(x,i)); i++) compressorWrite(self,t,vm);
	 return;
      }
      case IntType:
	 c = (Byte)convertToInt(x,vm);	// lowest byte
	 ptr = &c; size = 1; 
	 break;
      default:
      {
	 char	buf[100];
	 sprintf(buf,"ZLib.Compressor.write: Can't handle %s",iname(x));
	 vmThrow(vm,E_TYPE_ERROR,buf);
      }
   }
   if (size)
   {
      spinLockAcquire(Z_LOCK(self));
         _compress(self,ptr,size,vm);
      spinLockRelease(Z_LOCK(self));
   }
}

    // Compressor.write()
static Instance *Compressor_write(Instance *self,pArglist arglist,pVM vm)
{
   Instance *i;
   int	     n;

   for (n = 0; (i = listGet(arglist,n)); n++) compressorWrite(self,i,vm);
   if (n) asyncWriteHappened(self);

   return self;
}

static MethodTable compressorMethods[] =
{
   "create",		(pMethod)Compressor_open,
   "open",		(pMethod)Compressor_open,
   "close",		(pMethod)ZLib_close,
   "flush",		(pMethod)Object_noop,
   "write",		(pMethod)Compressor_write,
   "read",		(pMethod)ZLib_read,
   "drain",		(pMethod)ZLib_drain,
   "len",		(pMethod)ZLib_len,
   "pump",		(pMethod)ZLib_pump,
   "walker",		(pMethod)ZLib_walker,
   "calcAdler32",	(pMethod)calcAdler32,
   "calcCRC32",		(pMethod)calcCRC32,
   0,			0
};


/* ******************************************************************** */
/* ************************* Inflator Methods ************************* */
/* ******************************************************************** */

    // Inflator.open([use gzip header]), Inflator.create()
    // NOTE! if you open an existing Inflator, you still get a new one
static Instance *Inflator_open(Instance *self,pArglist arglist,pVM vm)
   { return _open(INFLATOR,arglist,vm); }

    // Inflator.write(data)
static Instance *Inflator_write(Instance *self,pArglist arglist,pVM vm)
{
   size_t    size;
   Instance *data = arglistGetBT(arglist,0,DataType,"Inflator.write",vm);
   void	    *ptr  = dataText(data,&size);

   if (size) _uncompress(self,ptr,size,vm);

   return self;
}


static MethodTable inflatorMethods[] =
{
   "create",		(pMethod)Inflator_open,
   "open",		(pMethod)Inflator_open,
   "close",		(pMethod)ZLib_close,
   "flush",		(pMethod)Object_noop,
   "write",		(pMethod)Inflator_write,
   "read",		(pMethod)ZLib_read,
   "drain",		(pMethod)ZLib_drain,
   "len",		(pMethod)ZLib_len,
   "pump",		(pMethod)ZLib_pump,
   "walker",		(pMethod)ZLib_walker,
   "calcAdler32",	(pMethod)calcAdler32,
   "calcCRC32",		(pMethod)calcCRC32,
   0,			0
};

/* ******************************************************************** */
/* *************************** Properties ***************************** */
/* ******************************************************************** */

    // *.version
static Instance *ZLib_version(Instance *self, pVM vm)
   { return stringCreate((char *)zlibVersion(),I_OWNED,vm); }

    // *.chunkSize
static Instance *ZLib_chunkSize(Instance *self, pVM vm)
   { return intCreate(CHUNK,vm); }

    // *.authors
static Instance *authors(Instance *self, pVM vm)
{
   return createTupleOfStrings(vm,"Craig Durland","http://zenkinetic.com/",
          (char *)0); 
}


    // *.libraries
static Instance *libraries(Instance *self, pVM vm)
{
   return createTupleOfStrings(vm,
      "zlib library: Jean-loup Gailly and Mark Adler","http://zlib.net/",
      (char *)0);
}

    // [Compressor|Inflator].isClosed
static Instance *ZLib_isClosed(Instance *self,pVM vm)
   { return boolCreate(!Z_OPEN(self)); }

    // [Compressor|Inflator].adler32
static Instance *ZLib_adler32(Instance *self,pVM vm)
   { return intCreate(Z_STREAM(self)->adler,vm); }

static PropertyTable zproperties[] = 
{
   "version",		(pProperty)ZLib_version,
   "chunkSize",		(pProperty)ZLib_chunkSize,
   "authors",		(pProperty)authors,
   "libraries",		(pProperty)libraries,
   0,			0
};

static PropertyTable properties[] = 
{
   "isClosed",		(pProperty)ZLib_isClosed,
   "version",		(pProperty)ZLib_version,
   "chunkSize",		(pProperty)ZLib_chunkSize,
   "authors",		(pProperty)authors,
   "libraries",		(pProperty)libraries,
   "adler32",		(pProperty)ZLib_adler32,
   0,			0
};

/* ******************************************************************** */
/* ***************************** ZeeLib ******************************* */
/* ******************************************************************** */

static ZKL_Object ZeeLibObject;
static Instance	  ZeeLibInstance;

    // ZeeLib.calcAdler32(data)
static Instance *calcAdler32(Instance *self,pArglist arglist,pVM vm)
{
   Instance    *data = arglistGetBT(arglist,0,DataType,"calcAdler32",vm);
   uLong 	checksum;
   size_t	srcSize;
   Byte	       *src = dataText(data,&srcSize);

   checksum = adler32(0,0,0);
   checksum = adler32(checksum,src,(uInt)srcSize);

   return intCreate(checksum,vm);
}

    // ZeeLib.calcCRC32(data)
static Instance *calcCRC32(Instance *self,pArglist arglist,pVM vm)
{
   Instance    *data = arglistGetBT(arglist,0,DataType,"calcCRC32",vm);
   uLong 	checksum;
   size_t	srcSize;
   Byte	       *src = dataText(data,&srcSize);

   checksum = crc32(0,0,0);
   checksum = crc32(checksum,src,(uInt)srcSize);

   return intCreate(checksum,vm);
}

static MethodTable zeeLibMethods[] = 
{
   "Compressor",	(pMethod)Compressor_open,
   "Inflator",		(pMethod)Inflator_open,
   "calcAdler32",	(pMethod)calcAdler32,
   "calcCRC32",		(pMethod)calcCRC32,
   0,			0
};


#define Z_VAULT_PATH "Utils.Compression"

DllExport void *construct(void *vm)
{
   if (!vm) return (void *)ZKLX_PROTOCOL;	// handshake

   	// If this is reloaded, nothing happens except 
	// construct() is called again so don't reinitialize
   if (!ZeeLibObject.name)	// static items are initialized to zero
   {
      constructObject(&ZeeLibObject,NativeType, zeeLibMethods,zproperties,0,vm);
      ZeeLibObject.name	     = "ZeeLib";
      ZeeLibObject.vaultPath = Z_VAULT_PATH;
      ZeeLibObject.isize     = sizeof(ZeeLibInstance);
      instanceInit((Instance *)&ZeeLibInstance,&ZeeLibObject,I_UNTOUCHABLE);

      constructObject(&CompressorObject,NativeType,compressorMethods,properties,0,vm);
      constructObject(&InflatorObject,  NativeType,inflatorMethods,  properties,0,vm);
      CompressorObject.name	  = "Compressor";
      InflatorObject.name	  = "Inflator";
      CompressorObject.vaultPath  = 
        InflatorObject.vaultPath  = Z_VAULT_PATH ".ZeeLib";
      CompressorObject.isize      = InflatorObject.isize = sizeof(ZLib);
      CompressorObject.threadSafe = 1;
      InflatorObject.threadSafe   = 0;

      CompressorObject.magicMarker = InflatorObject.magicMarker = magicMarker;
      CompressorObject.freeMe      = InflatorObject.freeMe	= freeMe;
   }
   return &ZeeLibInstance;
}
